@@ -2,8 +2,6 @@
 Juju Client
 -----------
 
-Seriously Alpha. Works now, but API *will* change.
-
 A simple synchronous python client for the juju-core websocket api.
 
 Example Usage::
@@ -39,28 +37,26 @@ Todo
 - Provide a buffered in mem option with watches on a single conn.
 
 Upstream/Server
-  - need proper status output, or other introspection beyond AllWatcher
-  - deploy local charm
   - bad constraints fail silently
-
-  - need terminate machine api
-  - clarify usage/working of env annotation
+  - need proper charm api
+  -
 """
 # License: GPLv3
 # Author: Kapil Thangavelu <kapil.foss@gmail.com>
 
 from base64 import b64encode
 from contextlib import contextmanager
-
 import errno
 import httplib
 import json
+import logging
+import os
 import pprint
-import time
 import signal
 import socket
+import ssl
 import StringIO
-import logging
+import time
 import websocket
 
 # There are two pypi modules with the name websocket (python-websocket
@@ -175,18 +171,9 @@ class RPC(object):
             return False
 
         log.info("Reconnecting client")
-        while True:
-            try:
-                self.conn = websocket.create_connection(
-                    self._reconnect_params['url'],
-                    origin=self._reconnect_params['origin'])
-                break
-            except socket.error as err:
-                if not err.errno in (
-                        errno.ETIMEDOUT, errno.ECONNREFUSED, errno.ECONNRESET):
-                    raise
-                time.sleep(1)
-                continue
+        self.conn = Connector.connect_socket_loop(
+            self._reconnect_params['url'],
+            self._reconnect_params['ca_cert'])
         self.login(**self._reconnect_params)
         return True
 
@@ -292,6 +279,7 @@ class TimeoutWatcher(Watcher):
             signal.alarm(timeout)
             yield None
         finally:
+            signal.alarm(0)
             signal.signal(signal.SIGALRM, signal.SIG_DFL)
 
     @classmethod
@@ -299,19 +287,119 @@ class TimeoutWatcher(Watcher):
         raise TimeoutError()
 
 
+class Connector(object):
+    """Abstract out the details of connecting to state servers.
+
+    Covers
+    - finding state servers, credentials, certs for a named env.
+    - verifying state servers are listening
+    - connecting an environment or websocket to a state server.
+    """
+
+    retry_conn_errors = (errno.ETIMEDOUT, errno.ECONNREFUSED, errno.ECONNRESET)
+
+    def run(self, cls, env_name):
+        """Given an environment name, return an authenticated client to it."""
+        jhome, data = self.parse_env(env_name)
+        cert_path = self.write_ca(jhome, env_name, data)
+        address = self.get_state_server(data)
+        if not address:
+            return
+        return self.connect_env(
+            cls, address, data['user'], data['password'], cert_path)
+
+    def connect_env(self, cls, address, user, password, cert_path=None):
+        """Given environment info return an authenticated client to it."""
+        env = cls("wss://%s" % address, ca_cert=cert_path)
+        env.login(user="user-%s" % user, password=password)
+        return env
+
+    @classmethod
+    def connect_socket(cls, endpoint, cert_path=None):
+        """Return a websocket connection to an endpoint."""
+        sslopt = {'ssl_version': ssl.PROTOCOL_TLSv1}
+        if cert_path:
+            sslopt['ca_certs'] = cert_path
+            # ssl.match_hostname is broken for us, need to disable per
+            # https://github.com/liris/websocket-client/issues/105
+            # when that's available, we can just selectively disable
+            # the host name match, for now we have to disable cert
+            # checking :-(
+            sslopt['check_hostname'] = False
+        else:
+            sslopt['cert_reqs'] = ssl.CERT_NONE
+
+        return websocket.create_connection(
+            endpoint, origin=endpoint, sslopt=sslopt)
+
+    def connect_socket_loop(self, endpoint, cert_path=None, timeout=120):
+        """Retry websocket connections to an endpoint till its connected."""
+        t = time.time()
+        while (time.time() > t + timeout):
+            try:
+                return Connector.connect_socket(endpoint, cert_path)
+            except socket.error as err:
+                if not err.errno in self.retry_conn_errors:
+                    raise
+                time.sleep(1)
+                continue
+
+    def write_ca(self, cert_dir, cert_name, data):
+        """Write ssl ca to the given."""
+        cert_path = os.path.join(cert_dir, '%s-cacert.pem' % cert_name)
+        with open(cert_path, 'w') as ca_fh:
+            ca_fh.write(data['ca-cert'])
+        return cert_path
+
+    def get_state_server(self, data):
+        """Given a list of state servers, return one that's listening."""
+        found = False
+        for s in data['state-servers']:
+            if self.is_server_available(s):
+                found = True
+                break
+        if not found:
+            return
+        return s
+
+    def parse_env(self, env_name):
+        import yaml
+        jhome = os.path.expanduser(
+            os.environ.get('JUJU_HOME', '~/.juju'))
+        jenv = os.path.join(jhome, 'environments', '%s.jenv' % env_name)
+        if not os.path.exists(jenv):
+            raise ValueError("Environment %s not bootstrapped" % env_name)
+
+        with open(jenv) as fh:
+            data = yaml.safe_load(fh.read())
+            return jhome, data
+
+    def is_server_available(self, server):
+        """ Given address/port, return true/false if it's up """
+        address, port = server.split(":")
+        try:
+            socket.create_connection((address, port), 3)
+            return True
+        except socket.error as err:
+            if err.errno in self.retry_conn_errors:
+                return False
+            else:
+                raise
+
+
 class Environment(RPC):
 
-    def __init__(self, endpoint, conn=None):
+    def __init__(self, endpoint, conn=None, ca_cert=None):
         self.endpoint = endpoint
         self._watches = []
         # For watches.
         self._creds = None
+        self._ca_cert = ca_cert
 
         if conn is not None:
             self.conn = conn
         else:
-            self.conn = websocket.create_connection(
-                endpoint, origin=self.endpoint)
+            self.conn = Connector.connect_socket(endpoint, self._ca_cert)
 
     def close(self):
         for w in self._watches:
@@ -319,6 +407,10 @@ class Environment(RPC):
         if self.conn.connected:
             self.conn.close()
 
+    @classmethod
+    def connect(cls, env_name):
+        return Connector().run(cls, env_name)
+
     # Charm ops
     def add_local_charm(self, charm_file, series, size=None):
         """Add a local charm to an environment.
@@ -357,10 +449,13 @@ class Environment(RPC):
             "Type": "Client",
             "Request": "EnvironmentInfo"})
 
-    def status(self):
-        # Status via api is currently broken, only reports machine ids,
-        # use an all watch with status translator to get something usable.
-        return self._rpc({"Type": "Client", "Request": "Status"})
+    def status(self, filters=None):
+        if isinstance(filters, basestring):
+            filters = [filters]
+        op = {"Type": "Client", "Request": "FullStatus"}
+        if filters:
+            op["Params"] = {'Patterns': filters}
+        return self._rpc(op)
 
     def get_charm(self, charm_url):
         return self._rpc(
@@ -489,6 +584,14 @@ class Environment(RPC):
                 "Series": series,
                 "Arch": arch}})
 
+    def retry_provisioning(self, machines):
+        """Mark machines for provisioner to retry iaas provisioning."""
+        return self._rpc({
+            "Type": "Client",
+            "Request": "RetryProvisioning",
+            "Params": {
+                "Entities": map(lambda x: {"Tag": "machine-%s"}, machines)}})
+
     # Watch Wrapper methods
     def get_stat(self):
         """A status emulator using the watch api, returns immediately.
@@ -523,7 +626,10 @@ class Environment(RPC):
             watch_env = connection
 
         p = dict(self._creds)
-        p.update({'url': self.endpoint, 'origin': self.endpoint})
+        p.update({
+            'url': self.endpoint,
+            'origin': self.endpoint,
+            'ca_cert': self._ca_cert})
         if timeout is not None:
             if watch_class is None:
                 watch_class = TimeoutWatcher
@@ -665,7 +771,7 @@ class Environment(RPC):
         """Update a service.
 
         Can update a service's charm, modify configuration, constraints,
-        and the number of units.
+        and the minimum number of units.
         """
         svc_config = {}
         if settings:
@@ -673,7 +779,7 @@ class Environment(RPC):
 
         return self._rpc(
             {"Type": "Client",
-             "Request": "SetServiceConstraints",
+             "Request": "ServiceUpdate",
              "Params": {
                  "ServiceName": service_name,
                  "CharmUrl": charm_url,
@@ -944,8 +1050,7 @@ class StatusTranslator(object):
         for ep in d['Endpoints']:
             svc_rels = self.data.setdefault(
                 'services', {}).setdefault(
-                    ep['ServiceName'], {}).setdefault(
-                        'relations', {})
+                    ep['ServiceName'], {}).setdefault('relations', {})
             svc_rels.setdefault(
                 ep['Relation']['Name'], []).append(ep['RemoteService'])
 
